@@ -1,12 +1,12 @@
-#define PY_SSIZE_T_CLEAN
-#include <Python.h>
-#include "structmember.h"
-
 #define NPY_NO_DEPRECATED_API NPY_API_VERSION
 #define _MULTIARRAYMODULE
+
+#define PY_SSIZE_T_CLEAN
+#include <Python.h>
+#include <structmember.h>
+
 #include "numpy/arrayobject.h"
 #include "numpy/arrayscalars.h"
-#include "numpy/arrayobject.h"
 
 #include "npy_config.h"
 #include "npy_pycompat.h"
@@ -15,6 +15,13 @@
 #include "arraytypes.h"
 
 #include "conversion_utils.h"
+#include "alloc.h"
+#include "npy_buffer.h"
+
+static int
+PyArray_PyIntAsInt_ErrMsg(PyObject *o, const char * msg) NPY_GCC_NONNULL(2);
+static npy_intp
+PyArray_PyIntAsIntp_ErrMsg(PyObject *o, const char * msg) NPY_GCC_NONNULL(2);
 
 /****************************************************************
 * Useful function for conversion when used with PyArg_ParseTuple
@@ -40,8 +47,7 @@ PyArray_Converter(PyObject *object, PyObject **address)
         return NPY_SUCCEED;
     }
     else {
-        *address = PyArray_FromAny(object, NULL, 0, 0,
-                                NPY_ARRAY_CARRAY, NULL);
+        *address = PyArray_FROM_OF(object, NPY_ARRAY_CARRAY);
         if (*address == NULL) {
             return NPY_FAIL;
         }
@@ -89,9 +95,21 @@ PyArray_IntpConverter(PyObject *obj, PyArray_Dims *seq)
 
     seq->ptr = NULL;
     seq->len = 0;
+
+    /*
+     * When the deprecation below expires, remove the `if` statement, and
+     * update the comment for PyArray_OptionalIntpConverter.
+     */
     if (obj == Py_None) {
+        /* Numpy 1.20, 2020-05-31 */
+        if (DEPRECATE(
+                "Passing None into shape arguments as an alias for () is "
+                "deprecated.") < 0){
+            return NPY_FAIL;
+        }
         return NPY_SUCCEED;
     }
+
     len = PySequence_Size(obj);
     if (len == -1) {
         /* Check to see if it is an integer number */
@@ -110,12 +128,12 @@ PyArray_IntpConverter(PyObject *obj, PyArray_Dims *seq)
         return NPY_FAIL;
     }
     if (len > NPY_MAXDIMS) {
-        PyErr_Format(PyExc_ValueError, "sequence too large; "
-                     "must be smaller than %d", NPY_MAXDIMS);
+        PyErr_Format(PyExc_ValueError, "maximum supported dimension for an ndarray is %d"
+                     ", found %d", NPY_MAXDIMS, len);
         return NPY_FAIL;
     }
     if (len > 0) {
-        seq->ptr = PyDimMem_NEW(len);
+        seq->ptr = npy_alloc_cache_dim(len);
         if (seq->ptr == NULL) {
             PyErr_NoMemory();
             return NPY_FAIL;
@@ -124,13 +142,63 @@ PyArray_IntpConverter(PyObject *obj, PyArray_Dims *seq)
     seq->len = len;
     nd = PyArray_IntpFromIndexSequence(obj, (npy_intp *)seq->ptr, len);
     if (nd == -1 || nd != len) {
-        PyDimMem_FREE(seq->ptr);
+        npy_free_cache_dim_obj(*seq);
         seq->ptr = NULL;
         return NPY_FAIL;
     }
     return NPY_SUCCEED;
 }
 
+/*
+ * Like PyArray_IntpConverter, but leaves `seq` untouched if `None` is passed
+ * rather than treating `None` as `()`.
+ */
+NPY_NO_EXPORT int
+PyArray_OptionalIntpConverter(PyObject *obj, PyArray_Dims *seq)
+{
+    if (obj == Py_None) {
+        return NPY_SUCCEED;
+    }
+
+    return PyArray_IntpConverter(obj, seq);
+}
+
+NPY_NO_EXPORT int
+PyArray_CopyConverter(PyObject *obj, _PyArray_CopyMode *copymode) {
+    if (obj == Py_None) {
+        PyErr_SetString(PyExc_ValueError,
+                        "NoneType copy mode not allowed.");
+        return NPY_FAIL;
+    }
+
+    int int_copymode;
+    static PyObject* numpy_CopyMode = NULL;
+    npy_cache_import("numpy", "_CopyMode", &numpy_CopyMode);
+
+    if (numpy_CopyMode != NULL && (PyObject *)Py_TYPE(obj) == numpy_CopyMode) {
+        PyObject* mode_value = PyObject_GetAttrString(obj, "value");
+        if (mode_value == NULL) {
+            return NPY_FAIL;
+        }
+
+        int_copymode = (int)PyLong_AsLong(mode_value);
+        Py_DECREF(mode_value);
+        if (error_converting(int_copymode)) {
+            return NPY_FAIL;
+        }
+    }
+    else {
+        npy_bool bool_copymode;
+        if (!PyArray_BoolConverter(obj, &bool_copymode)) {
+            return NPY_FAIL;
+        }
+        int_copymode = (int)bool_copymode;
+    }
+
+    *copymode = (_PyArray_CopyMode)int_copymode;
+    return NPY_SUCCEED;
+}
+
 /*NUMPY_API
  * Get buffer chunk from object
  *
@@ -146,11 +214,7 @@ PyArray_IntpConverter(PyObject *obj, PyArray_Dims *seq)
 NPY_NO_EXPORT int
 PyArray_BufferConverter(PyObject *obj, PyArray_Chunk *buf)
 {
-#if defined(NPY_PY3K)
     Py_buffer view;
-#else
-    Py_ssize_t buflen;
-#endif
 
     buf->ptr = NULL;
     buf->flags = NPY_ARRAY_BEHAVED;
@@ -159,11 +223,12 @@ PyArray_BufferConverter(PyObject *obj, PyArray_Chunk *buf)
         return NPY_SUCCEED;
     }
 
-#if defined(NPY_PY3K)
-    if (PyObject_GetBuffer(obj, &view, PyBUF_ANY_CONTIGUOUS|PyBUF_WRITABLE) != 0) {
+    if (PyObject_GetBuffer(obj, &view,
+                PyBUF_ANY_CONTIGUOUS|PyBUF_WRITABLE|PyBUF_SIMPLE) != 0) {
         PyErr_Clear();
         buf->flags &= ~NPY_ARRAY_WRITEABLE;
-        if (PyObject_GetBuffer(obj, &view, PyBUF_ANY_CONTIGUOUS) != 0) {
+        if (PyObject_GetBuffer(obj, &view,
+                PyBUF_ANY_CONTIGUOUS|PyBUF_SIMPLE) != 0) {
             return NPY_FAIL;
         }
     }
@@ -172,8 +237,10 @@ PyArray_BufferConverter(PyObject *obj, PyArray_Chunk *buf)
     buf->len = (npy_intp) view.len;
 
     /*
-     * XXX: PyObject_AsWriteBuffer does also this, but it is unsafe, as there is
-     * no strict guarantee that the buffer sticks around after being released.
+     * In Python 3 both of the deprecated functions PyObject_AsWriteBuffer and
+     * PyObject_AsReadBuffer that this code replaces release the buffer. It is
+     * up to the object that supplies the buffer to guarantee that the buffer
+     * sticks around after the release.
      */
     PyBuffer_Release(&view);
 
@@ -181,22 +248,6 @@ PyArray_BufferConverter(PyObject *obj, PyArray_Chunk *buf)
     if (PyMemoryView_Check(obj)) {
         buf->base = PyMemoryView_GET_BASE(obj);
     }
-#else
-    if (PyObject_AsWriteBuffer(obj, &(buf->ptr), &buflen) < 0) {
-        PyErr_Clear();
-        buf->flags &= ~NPY_ARRAY_WRITEABLE;
-        if (PyObject_AsReadBuffer(obj, (const void **)&(buf->ptr),
-                                  &buflen) < 0) {
-            return NPY_FAIL;
-        }
-    }
-    buf->len = (npy_intp) buflen;
-
-    /* Point to the base of the buffer object if present */
-    if (PyBuffer_Check(obj)) {
-        buf->base = ((PyArray_Chunk *)obj)->base;
-    }
-#endif
     if (buf->base == NULL) {
         buf->base = obj;
     }
@@ -215,8 +266,9 @@ PyArray_AxisConverter(PyObject *obj, int *axis)
         *axis = NPY_MAXDIMS;
     }
     else {
-        *axis = PyArray_PyIntAsInt(obj);
-        if (PyErr_Occurred()) {
+        *axis = PyArray_PyIntAsInt_ErrMsg(obj,
+                               "an integer is required for the axis");
+        if (error_converting(*axis)) {
             return NPY_FAIL;
         }
     }
@@ -251,18 +303,12 @@ PyArray_ConvertMultiAxis(PyObject *axis_in, int ndim, npy_bool *out_axis_flags)
         }
         for (i = 0; i < naxes; ++i) {
             PyObject *tmp = PyTuple_GET_ITEM(axis_in, i);
-            int axis = PyArray_PyIntAsInt(tmp);
-            int axis_orig = axis;
+            int axis = PyArray_PyIntAsInt_ErrMsg(tmp,
+                          "integers are required for the axis tuple elements");
             if (error_converting(axis)) {
                 return NPY_FAIL;
             }
-            if (axis < 0) {
-                axis += ndim;
-            }
-            if (axis < 0 || axis >= ndim) {
-                PyErr_Format(PyExc_ValueError,
-                        "'axis' entry %d is out of bounds [-%d, %d)",
-                        axis_orig, ndim, ndim);
+            if (check_and_adjust_axis(&axis, ndim) < 0) {
                 return NPY_FAIL;
             }
             if (out_axis_flags[axis]) {
@@ -277,19 +323,16 @@ PyArray_ConvertMultiAxis(PyObject *axis_in, int ndim, npy_bool *out_axis_flags)
     }
     /* Try to interpret axis as an integer */
     else {
-        int axis, axis_orig;
+        int axis;
 
         memset(out_axis_flags, 0, ndim);
 
-        axis = PyArray_PyIntAsInt(axis_in);
-        axis_orig = axis;
+        axis = PyArray_PyIntAsInt_ErrMsg(axis_in,
+                                   "an integer is required for the axis");
 
         if (error_converting(axis)) {
             return NPY_FAIL;
         }
-        if (axis < 0) {
-            axis += ndim;
-        }
         /*
          * Special case letting axis={-1,0} slip through for scalars,
          * for backwards compatibility reasons.
@@ -298,10 +341,7 @@ PyArray_ConvertMultiAxis(PyObject *axis_in, int ndim, npy_bool *out_axis_flags)
             return NPY_SUCCEED;
         }
 
-        if (axis < 0 || axis >= ndim) {
-            PyErr_Format(PyExc_ValueError,
-                    "'axis' entry %d is out of bounds [-%d, %d)",
-                    axis_orig, ndim, ndim);
+        if (check_and_adjust_axis(&axis, ndim) < 0) {
             return NPY_FAIL;
         }
 
@@ -329,59 +369,140 @@ PyArray_BoolConverter(PyObject *object, npy_bool *val)
     return NPY_SUCCEED;
 }
 
+static int
+string_converter_helper(
+    PyObject *object,
+    void *out,
+    int (*str_func)(char const*, Py_ssize_t, void*),
+    char const *name,
+    char const *message)
+{
+    /* allow bytes for compatibility */
+    PyObject *str_object = NULL;
+    if (PyBytes_Check(object)) {
+        str_object = PyUnicode_FromEncodedObject(object, NULL, NULL);
+        if (str_object == NULL) {
+            PyErr_Format(PyExc_ValueError,
+                "%s %s (got %R)", name, message, object);
+            return NPY_FAIL;
+        }
+    }
+    else if (PyUnicode_Check(object)) {
+        str_object = object;
+        Py_INCREF(str_object);
+    }
+    else {
+        PyErr_Format(PyExc_TypeError,
+            "%s must be str, not %s", name, Py_TYPE(object)->tp_name);
+        return NPY_FAIL;
+    }
+
+    Py_ssize_t length;
+    char const *str = PyUnicode_AsUTF8AndSize(str_object, &length);
+    if (str == NULL) {
+        Py_DECREF(str_object);
+        return NPY_FAIL;
+    }
+
+    int ret = str_func(str, length, out);
+    Py_DECREF(str_object);
+    if (ret < 0) {
+        /* str_func returns -1 without an exception if the value is wrong */
+        if (!PyErr_Occurred()) {
+            PyErr_Format(PyExc_ValueError,
+                "%s %s (got %R)", name, message, object);
+        }
+        return NPY_FAIL;
+    }
+    return NPY_SUCCEED;
+}
+
+static int byteorder_parser(char const *str, Py_ssize_t length, void *data)
+{
+    char *endian = (char *)data;
+
+    if (length < 1) {
+        return -1;
+    }
+    else if (str[0] == NPY_BIG || str[0] == NPY_LITTLE ||
+             str[0] == NPY_NATIVE || str[0] == NPY_IGNORE) {
+        *endian = str[0];
+        return 0;
+    }
+    else if (str[0] == 'b' || str[0] == 'B') {
+        *endian = NPY_BIG;
+        return 0;
+    }
+    else if (str[0] == 'l' || str[0] == 'L') {
+        *endian = NPY_LITTLE;
+        return 0;
+    }
+    else if (str[0] == 'n' || str[0] == 'N') {
+        *endian = NPY_NATIVE;
+        return 0;
+    }
+    else if (str[0] == 'i' || str[0] == 'I') {
+        *endian = NPY_IGNORE;
+        return 0;
+    }
+    else if (str[0] == 's' || str[0] == 'S') {
+        *endian = NPY_SWAP;
+        return 0;
+    }
+    else {
+        return -1;
+    }
+}
+
 /*NUMPY_API
  * Convert object to endian
  */
 NPY_NO_EXPORT int
 PyArray_ByteorderConverter(PyObject *obj, char *endian)
 {
-    char *str;
-    PyObject *tmp = NULL;
+    return string_converter_helper(
+        obj, (void *)endian, byteorder_parser, "byteorder", "not recognized");
+}
 
-    if (PyUnicode_Check(obj)) {
-        obj = tmp = PyUnicode_AsASCIIString(obj);
-    }
+static int sortkind_parser(char const *str, Py_ssize_t length, void *data)
+{
+    NPY_SORTKIND *sortkind = (NPY_SORTKIND *)data;
 
-    *endian = NPY_SWAP;
-    str = PyBytes_AsString(obj);
-    if (!str) {
-        Py_XDECREF(tmp);
-        return NPY_FAIL;
+    if (length < 1) {
+        return -1;
     }
-    if (strlen(str) < 1) {
-        PyErr_SetString(PyExc_ValueError,
-                        "Byteorder string must be at least length 1");
-        Py_XDECREF(tmp);
-        return NPY_FAIL;
+    if (str[0] == 'q' || str[0] == 'Q') {
+        *sortkind = NPY_QUICKSORT;
+        return 0;
     }
-    *endian = str[0];
-    if (str[0] != NPY_BIG && str[0] != NPY_LITTLE
-        && str[0] != NPY_NATIVE && str[0] != NPY_IGNORE) {
-        if (str[0] == 'b' || str[0] == 'B') {
-            *endian = NPY_BIG;
-        }
-        else if (str[0] == 'l' || str[0] == 'L') {
-            *endian = NPY_LITTLE;
-        }
-        else if (str[0] == 'n' || str[0] == 'N') {
-            *endian = NPY_NATIVE;
-        }
-        else if (str[0] == 'i' || str[0] == 'I') {
-            *endian = NPY_IGNORE;
-        }
-        else if (str[0] == 's' || str[0] == 'S') {
-            *endian = NPY_SWAP;
-        }
-        else {
-            PyErr_Format(PyExc_ValueError,
-                         "%s is an unrecognized byteorder",
-                         str);
-            Py_XDECREF(tmp);
-            return NPY_FAIL;
-        }
+    else if (str[0] == 'h' || str[0] == 'H') {
+        *sortkind = NPY_HEAPSORT;
+        return 0;
+    }
+    else if (str[0] == 'm' || str[0] == 'M') {
+        /*
+         * Mergesort is an alias for NPY_STABLESORT.
+         * That maintains backwards compatibility while
+         * allowing other types of stable sorts to be used.
+         */
+        *sortkind = NPY_MERGESORT;
+        return 0;
+    }
+    else if (str[0] == 's' || str[0] == 'S') {
+        /*
+         * NPY_STABLESORT is one of
+         *
+         *   - mergesort
+         *   - timsort
+         *
+         *  Which one is used depends on the data type.
+         */
+        *sortkind = NPY_STABLESORT;
+        return 0;
+    }
+    else {
+        return -1;
     }
-    Py_XDECREF(tmp);
-    return NPY_SUCCEED;
 }
 
 /*NUMPY_API
@@ -390,43 +511,26 @@ PyArray_ByteorderConverter(PyObject *obj, char *endian)
 NPY_NO_EXPORT int
 PyArray_SortkindConverter(PyObject *obj, NPY_SORTKIND *sortkind)
 {
-    char *str;
-    PyObject *tmp = NULL;
+    /* Leave the desired default from the caller for Py_None */
+    if (obj == Py_None) {
+        return NPY_SUCCEED;
+    }
+    return string_converter_helper(
+        obj, (void *)sortkind, sortkind_parser, "sort kind",
+        "must be one of 'quick', 'heap', or 'stable'");
+}
 
-    if (PyUnicode_Check(obj)) {
-        obj = tmp = PyUnicode_AsASCIIString(obj);
-    }
+static int selectkind_parser(char const *str, Py_ssize_t length, void *data)
+{
+    NPY_SELECTKIND *selectkind = (NPY_SELECTKIND *)data;
 
-    *sortkind = NPY_QUICKSORT;
-    str = PyBytes_AsString(obj);
-    if (!str) {
-        Py_XDECREF(tmp);
-        return NPY_FAIL;
-    }
-    if (strlen(str) < 1) {
-        PyErr_SetString(PyExc_ValueError,
-                        "Sort kind string must be at least length 1");
-        Py_XDECREF(tmp);
-        return NPY_FAIL;
-    }
-    if (str[0] == 'q' || str[0] == 'Q') {
-        *sortkind = NPY_QUICKSORT;
-    }
-    else if (str[0] == 'h' || str[0] == 'H') {
-        *sortkind = NPY_HEAPSORT;
-    }
-    else if (str[0] == 'm' || str[0] == 'M') {
-        *sortkind = NPY_MERGESORT;
+    if (length == 11 && strcmp(str, "introselect") == 0) {
+        *selectkind = NPY_INTROSELECT;
+        return 0;
     }
     else {
-        PyErr_Format(PyExc_ValueError,
-                     "%s is an unrecognized kind of sort",
-                     str);
-        Py_XDECREF(tmp);
-        return NPY_FAIL;
+        return -1;
     }
-    Py_XDECREF(tmp);
-    return NPY_SUCCEED;
 }
 
 /*NUMPY_API
@@ -435,37 +539,44 @@ PyArray_SortkindConverter(PyObject *obj, NPY_SORTKIND *sortkind)
 NPY_NO_EXPORT int
 PyArray_SelectkindConverter(PyObject *obj, NPY_SELECTKIND *selectkind)
 {
-    char *str;
-    PyObject *tmp = NULL;
+    return string_converter_helper(
+        obj, (void *)selectkind, selectkind_parser, "select kind",
+        "must be 'introselect'");
+}
 
-    if (PyUnicode_Check(obj)) {
-        obj = tmp = PyUnicode_AsASCIIString(obj);
-    }
+static int searchside_parser(char const *str, Py_ssize_t length, void *data)
+{
+    NPY_SEARCHSIDE *side = (NPY_SEARCHSIDE *)data;
+    int is_exact = 0;
 
-    *selectkind = NPY_INTROSELECT;
-    str = PyBytes_AsString(obj);
-    if (!str) {
-        Py_XDECREF(tmp);
-        return NPY_FAIL;
+    if (length < 1) {
+        return -1;
     }
-    if (strlen(str) < 1) {
-        PyErr_SetString(PyExc_ValueError,
-                        "Select kind string must be at least length 1");
-        Py_XDECREF(tmp);
-        return NPY_FAIL;
+    else if (str[0] == 'l' || str[0] == 'L') {
+        *side = NPY_SEARCHLEFT;
+        is_exact = (length == 4 && strcmp(str, "left") == 0);
     }
-    if (strcmp(str, "introselect") == 0) {
-        *selectkind = NPY_INTROSELECT;
+    else if (str[0] == 'r' || str[0] == 'R') {
+        *side = NPY_SEARCHRIGHT;
+        is_exact = (length == 5 && strcmp(str, "right") == 0);
     }
     else {
-        PyErr_Format(PyExc_ValueError,
-                     "%s is an unrecognized kind of select",
-                     str);
-        Py_XDECREF(tmp);
-        return NPY_FAIL;
+        return -1;
     }
-    Py_XDECREF(tmp);
-    return NPY_SUCCEED;
+
+    /* Filters out the case sensitive/non-exact
+     * match inputs and other inputs and outputs DeprecationWarning
+     */
+    if (!is_exact) {
+        /* NumPy 1.20, 2020-05-19 */
+        if (DEPRECATE("inexact matches and case insensitive matches "
+                      "for search side are deprecated, please use "
+                      "one of 'left' or 'right' instead.") < 0) {
+            return -1;
+        }
+    }
+
+    return 0;
 }
 
 /*NUMPY_API
@@ -474,36 +585,36 @@ PyArray_SelectkindConverter(PyObject *obj, NPY_SELECTKIND *selectkind)
 NPY_NO_EXPORT int
 PyArray_SearchsideConverter(PyObject *obj, void *addr)
 {
-    NPY_SEARCHSIDE *side = (NPY_SEARCHSIDE *)addr;
-    char *str;
-    PyObject *tmp = NULL;
+    return string_converter_helper(
+        obj, addr, searchside_parser, "search side",
+        "must be 'left' or 'right'");
+}
 
-    if (PyUnicode_Check(obj)) {
-        obj = tmp = PyUnicode_AsASCIIString(obj);
+static int order_parser(char const *str, Py_ssize_t length, void *data)
+{
+    NPY_ORDER *val = (NPY_ORDER *)data;
+    if (length != 1) {
+        return -1;
     }
-
-    str = PyBytes_AsString(obj);
-    if (!str || strlen(str) < 1) {
-        PyErr_SetString(PyExc_ValueError,
-                        "expected nonempty string for keyword 'side'");
-        Py_XDECREF(tmp);
-        return NPY_FAIL;
+    if (str[0] == 'C' || str[0] == 'c') {
+        *val = NPY_CORDER;
+        return 0;
     }
-
-    if (str[0] == 'l' || str[0] == 'L') {
-        *side = NPY_SEARCHLEFT;
+    else if (str[0] == 'F' || str[0] == 'f') {
+        *val = NPY_FORTRANORDER;
+        return 0;
     }
-    else if (str[0] == 'r' || str[0] == 'R') {
-        *side = NPY_SEARCHRIGHT;
+    else if (str[0] == 'A' || str[0] == 'a') {
+        *val = NPY_ANYORDER;
+        return 0;
+    }
+    else if (str[0] == 'K' || str[0] == 'k') {
+        *val = NPY_KEEPORDER;
+        return 0;
     }
     else {
-        PyErr_Format(PyExc_ValueError,
-                     "'%s' is an invalid value for keyword 'side'", str);
-        Py_XDECREF(tmp);
-        return NPY_FAIL;
+        return -1;
     }
-    Py_XDECREF(tmp);
-    return NPY_SUCCEED;
 }
 
 /*NUMPY_API
@@ -512,52 +623,52 @@ PyArray_SearchsideConverter(PyObject *obj, void *addr)
 NPY_NO_EXPORT int
 PyArray_OrderConverter(PyObject *object, NPY_ORDER *val)
 {
-    char *str;
-    /* Leave the desired default from the caller for NULL/Py_None */
-    if (object == NULL || object == Py_None) {
+    /* Leave the desired default from the caller for Py_None */
+    if (object == Py_None) {
         return NPY_SUCCEED;
     }
-    else if (PyUnicode_Check(object)) {
-        PyObject *tmp;
-        int ret;
-        tmp = PyUnicode_AsASCIIString(object);
-        ret = PyArray_OrderConverter(tmp, val);
-        Py_DECREF(tmp);
-        return ret;
+    return string_converter_helper(
+        object, (void *)val, order_parser, "order",
+        "must be one of 'C', 'F', 'A', or 'K'");
+}
+
+static int clipmode_parser(char const *str, Py_ssize_t length, void *data)
+{
+    NPY_CLIPMODE *val = (NPY_CLIPMODE *)data;
+    int is_exact = 0;
+
+    if (length < 1) {
+        return -1;
     }
-    else if (!PyBytes_Check(object) || PyBytes_GET_SIZE(object) < 1) {
-        if (PyObject_IsTrue(object)) {
-            *val = NPY_FORTRANORDER;
-        }
-        else {
-            *val = NPY_CORDER;
-        }
-        if (PyErr_Occurred()) {
-            return NPY_FAIL;
-        }
-        return NPY_SUCCEED;
+    if (str[0] == 'C' || str[0] == 'c') {
+        *val = NPY_CLIP;
+        is_exact = (length == 4 && strcmp(str, "clip") == 0);
+    }
+    else if (str[0] == 'W' || str[0] == 'w') {
+        *val = NPY_WRAP;
+        is_exact = (length == 4 && strcmp(str, "wrap") == 0);
+    }
+    else if (str[0] == 'R' || str[0] == 'r') {
+        *val = NPY_RAISE;
+        is_exact = (length == 5 && strcmp(str, "raise") == 0);
     }
     else {
-        str = PyBytes_AS_STRING(object);
-        if (str[0] == 'C' || str[0] == 'c') {
-            *val = NPY_CORDER;
-        }
-        else if (str[0] == 'F' || str[0] == 'f') {
-            *val = NPY_FORTRANORDER;
-        }
-        else if (str[0] == 'A' || str[0] == 'a') {
-            *val = NPY_ANYORDER;
-        }
-        else if (str[0] == 'K' || str[0] == 'k') {
-            *val = NPY_KEEPORDER;
-        }
-        else {
-            PyErr_SetString(PyExc_TypeError,
-                            "order not understood");
-            return NPY_FAIL;
+        return -1;
+    }
+
+    /* Filters out the case sensitive/non-exact
+     * match inputs and other inputs and outputs DeprecationWarning
+     */
+    if (!is_exact) {
+        /* Numpy 1.20, 2020-05-19 */
+        if (DEPRECATE("inexact matches and case insensitive matches "
+                      "for clip mode are deprecated, please use "
+                      "one of 'clip', 'raise', or 'wrap' instead.") < 0) {
+            return -1;
         }
     }
-    return NPY_SUCCEED;
+
+    return 0;
 }
 
 /*NUMPY_API
@@ -569,33 +680,14 @@ PyArray_ClipmodeConverter(PyObject *object, NPY_CLIPMODE *val)
     if (object == NULL || object == Py_None) {
         *val = NPY_RAISE;
     }
-    else if (PyBytes_Check(object)) {
-        char *str;
-        str = PyBytes_AS_STRING(object);
-        if (str[0] == 'C' || str[0] == 'c') {
-            *val = NPY_CLIP;
-        }
-        else if (str[0] == 'W' || str[0] == 'w') {
-            *val = NPY_WRAP;
-        }
-        else if (str[0] == 'R' || str[0] == 'r') {
-            *val = NPY_RAISE;
-        }
-        else {
-            PyErr_SetString(PyExc_TypeError,
-                            "clipmode not understood");
-            return NPY_FAIL;
-        }
-    }
-    else if (PyUnicode_Check(object)) {
-        PyObject *tmp;
-        int ret;
-        tmp = PyUnicode_AsASCIIString(object);
-        ret = PyArray_ClipmodeConverter(tmp, val);
-        Py_DECREF(tmp);
-        return ret;
+
+    else if (PyBytes_Check(object) || PyUnicode_Check(object)) {
+        return string_converter_helper(
+            object, (void *)val, clipmode_parser, "clipmode",
+            "must be one of 'clip', 'raise', or 'wrap'");
     }
     else {
+        /* For users passing `np.RAISE`, `np.WRAP`, `np.CLIP` */
         int number = PyArray_PyIntAsInt(object);
         if (error_converting(number)) {
             goto fail;
@@ -605,7 +697,8 @@ PyArray_ClipmodeConverter(PyObject *object, NPY_CLIPMODE *val)
             *val = (NPY_CLIPMODE) number;
         }
         else {
-            goto fail;
+            PyErr_Format(PyExc_ValueError,
+                    "integer clipmode must be np.RAISE, np.WRAP, or np.CLIP");
         }
     }
     return NPY_SUCCEED;
@@ -629,8 +722,8 @@ PyArray_ConvertClipmodeSequence(PyObject *object, NPY_CLIPMODE *modes, int n)
     if (object && (PyTuple_Check(object) || PyList_Check(object))) {
         if (PySequence_Size(object) != n) {
             PyErr_Format(PyExc_ValueError,
-                    "list of clipmodes has wrong length (%d instead of %d)",
-                    (int)PySequence_Size(object), n);
+                    "list of clipmodes has wrong length (%zd instead of %d)",
+                    PySequence_Size(object), n);
             return NPY_FAIL;
         }
 
@@ -659,66 +752,128 @@ PyArray_ConvertClipmodeSequence(PyObject *object, NPY_CLIPMODE *modes, int n)
     return NPY_SUCCEED;
 }
 
+static int correlatemode_parser(char const *str, Py_ssize_t length, void *data)
+{
+    NPY_CORRELATEMODE *val = (NPY_CORRELATEMODE *)data;
+    int is_exact = 0;
+
+    if (length < 1) {
+        return -1;
+    }
+    if (str[0] == 'V' || str[0] == 'v') {
+        *val = NPY_VALID;
+        is_exact = (length == 5 && strcmp(str, "valid") == 0);
+    }
+    else if (str[0] == 'S' || str[0] == 's') {
+        *val = NPY_SAME;
+        is_exact = (length == 4 && strcmp(str, "same") == 0);
+    }
+    else if (str[0] == 'F' || str[0] == 'f') {
+        *val = NPY_FULL;
+        is_exact = (length == 4 && strcmp(str, "full") == 0);
+    }
+    else {
+        return -1;
+    }
+
+    /* Filters out the case sensitive/non-exact
+     * match inputs and other inputs and outputs DeprecationWarning
+     */
+    if (!is_exact) {
+        /* Numpy 1.21, 2021-01-19 */
+        if (DEPRECATE("inexact matches and case insensitive matches for "
+                      "convolve/correlate mode are deprecated, please "
+                      "use one of 'valid', 'same', or 'full' instead.") < 0) {
+            return -1;
+        }
+    }
+
+    return 0;
+}
+
+/*
+ * Convert an object to NPY_VALID / NPY_SAME / NPY_FULL
+ */
+NPY_NO_EXPORT int
+PyArray_CorrelatemodeConverter(PyObject *object, NPY_CORRELATEMODE *val)
+{
+    if (PyUnicode_Check(object)) {
+        return string_converter_helper(
+            object, (void *)val, correlatemode_parser, "mode",
+            "must be one of 'valid', 'same', or 'full'");
+    }
+
+    else {
+        /* For users passing integers */
+        int number = PyArray_PyIntAsInt(object);
+        if (error_converting(number)) {
+            PyErr_SetString(PyExc_TypeError,
+                        "convolve/correlate mode not understood");
+            return NPY_FAIL;
+        }
+        if (number <= (int) NPY_FULL
+                && number >= (int) NPY_VALID) {
+            *val = (NPY_CORRELATEMODE) number;
+            return NPY_SUCCEED;
+        }
+        else {
+            PyErr_Format(PyExc_ValueError,
+                    "integer convolve/correlate mode must be 0, 1, or 2");
+            return NPY_FAIL;
+        }
+    }
+}
+
+static int casting_parser(char const *str, Py_ssize_t length, void *data)
+{
+    NPY_CASTING *casting = (NPY_CASTING *)data;
+    if (length < 2) {
+        return -1;
+    }
+    switch (str[2]) {
+    case 0:
+        if (length == 2 && strcmp(str, "no") == 0) {
+            *casting = NPY_NO_CASTING;
+            return 0;
+        }
+        break;
+    case 'u':
+        if (length == 5 && strcmp(str, "equiv") == 0) {
+            *casting = NPY_EQUIV_CASTING;
+            return 0;
+        }
+        break;
+    case 'f':
+        if (length == 4 && strcmp(str, "safe") == 0) {
+            *casting = NPY_SAFE_CASTING;
+            return 0;
+        }
+        break;
+    case 'm':
+        if (length == 9 && strcmp(str, "same_kind") == 0) {
+            *casting = NPY_SAME_KIND_CASTING;
+            return 0;
+        }
+        break;
+    case 's':
+        if (length == 6 && strcmp(str, "unsafe") == 0) {
+            *casting = NPY_UNSAFE_CASTING;
+            return 0;
+        }
+        break;
+    }
+    return -1;
+}
+
 /*NUMPY_API
  * Convert any Python object, *obj*, to an NPY_CASTING enum.
  */
 NPY_NO_EXPORT int
 PyArray_CastingConverter(PyObject *obj, NPY_CASTING *casting)
 {
-    char *str = NULL;
-    Py_ssize_t length = 0;
-
-    if (PyUnicode_Check(obj)) {
-        PyObject *str_obj;
-        int ret;
-        str_obj = PyUnicode_AsASCIIString(obj);
-        if (str_obj == NULL) {
-            return 0;
-        }
-        ret = PyArray_CastingConverter(str_obj, casting);
-        Py_DECREF(str_obj);
-        return ret;
-    }
-
-    if (PyBytes_AsStringAndSize(obj, &str, &length) == -1) {
-        return 0;
-    }
-
-    if (length >= 2) switch (str[2]) {
-        case 0:
-            if (strcmp(str, "no") == 0) {
-                *casting = NPY_NO_CASTING;
-                return 1;
-            }
-            break;
-        case 'u':
-            if (strcmp(str, "equiv") == 0) {
-                *casting = NPY_EQUIV_CASTING;
-                return 1;
-            }
-            break;
-        case 'f':
-            if (strcmp(str, "safe") == 0) {
-                *casting = NPY_SAFE_CASTING;
-                return 1;
-            }
-            break;
-        case 'm':
-            if (strcmp(str, "same_kind") == 0) {
-                *casting = NPY_SAME_KIND_CASTING;
-                return 1;
-            }
-            break;
-        case 's':
-            if (strcmp(str, "unsafe") == 0) {
-                *casting = NPY_UNSAFE_CASTING;
-                return 1;
-            }
-            break;
-    }
-
-    PyErr_SetString(PyExc_ValueError,
-            "casting must be one of 'no', 'equiv', 'safe', "
+    return string_converter_helper(
+        obj, (void *)casting, casting_parser, "casting",
+            "must be one of 'no', 'equiv', 'safe', "
             "'same_kind', or 'unsafe'");
     return 0;
 }
@@ -727,13 +882,12 @@ PyArray_CastingConverter(PyObject *obj, NPY_CASTING *casting)
 * Other conversion functions
 *****************************/
 
-/*NUMPY_API*/
-NPY_NO_EXPORT int
-PyArray_PyIntAsInt(PyObject *o)
+static int
+PyArray_PyIntAsInt_ErrMsg(PyObject *o, const char * msg)
 {
     npy_intp long_value;
     /* This assumes that NPY_SIZEOF_INTP >= NPY_SIZEOF_INT */
-    long_value = PyArray_PyIntAsIntp(o);
+    long_value = PyArray_PyIntAsIntp_ErrMsg(o, msg);
 
 #if (NPY_SIZEOF_INTP > NPY_SIZEOF_INT)
     if ((long_value < INT_MIN) || (long_value > INT_MAX)) {
@@ -745,8 +899,14 @@ PyArray_PyIntAsInt(PyObject *o)
 }
 
 /*NUMPY_API*/
-NPY_NO_EXPORT npy_intp
-PyArray_PyIntAsIntp(PyObject *o)
+NPY_NO_EXPORT int
+PyArray_PyIntAsInt(PyObject *o)
+{
+    return PyArray_PyIntAsInt_ErrMsg(o, "an integer is required");
+}
+
+static npy_intp
+PyArray_PyIntAsIntp_ErrMsg(PyObject *o, const char * msg)
 {
 #if (NPY_SIZEOF_LONG < NPY_SIZEOF_INTP)
     long long long_value = -1;
@@ -754,38 +914,22 @@ PyArray_PyIntAsIntp(PyObject *o)
     long long_value = -1;
 #endif
     PyObject *obj, *err;
-    static char *msg = "an integer is required";
 
-    if (!o) {
+    /*
+     * Be a bit stricter and not allow bools.
+     * np.bool_ is also disallowed as Boolean arrays do not currently
+     * support index.
+     */
+    if (!o || PyBool_Check(o) || PyArray_IsScalar(o, Bool)) {
         PyErr_SetString(PyExc_TypeError, msg);
         return -1;
     }
 
-    /* Be a bit stricter and not allow bools, np.bool_ is handled later */
-    if (PyBool_Check(o)) {
-        if (DEPRECATE("using a boolean instead of an integer"
-                      " will result in an error in the future") < 0) {
-            return -1;
-        }
-    }
-
     /*
      * Since it is the usual case, first check if o is an integer. This is
      * an exact check, since otherwise __index__ is used.
      */
-#if !defined(NPY_PY3K)
-    if PyInt_CheckExact(o) {
-  #if (NPY_SIZEOF_LONG <= NPY_SIZEOF_INTP)
-        /* No overflow is possible, so we can just return */
-        return PyInt_AS_LONG(o);
-  #else
-        long_value = PyInt_AS_LONG(o);
-        goto overflow_check;
-  #endif
-    }
-    else
-#endif
-    if PyLong_CheckExact(o) {
+    if (PyLong_CheckExact(o)) {
 #if (NPY_SIZEOF_LONG < NPY_SIZEOF_INTP)
         long_value = PyLong_AsLongLong(o);
 #else
@@ -794,82 +938,22 @@ PyArray_PyIntAsIntp(PyObject *o)
         return (npy_intp)long_value;
     }
 
-    /* Disallow numpy.bool_. Boolean arrays do not currently support index. */
-    if (PyArray_IsScalar(o, Bool)) {
-        if (DEPRECATE("using a boolean instead of an integer"
-                      " will result in an error in the future") < 0) {
-            return -1;
-        }
-    }
-
     /*
      * The most general case. PyNumber_Index(o) covers everything
      * including arrays. In principle it may be possible to replace
      * the whole function by PyIndex_AsSSize_t after deprecation.
      */
     obj = PyNumber_Index(o);
-    if (obj) {
-#if (NPY_SIZEOF_LONG < NPY_SIZEOF_INTP)
-        long_value = PyLong_AsLongLong(obj);
-#else
-        long_value = PyLong_AsLong(obj);
-#endif
-        Py_DECREF(obj);
-        goto finish;
-    }
-    else {
-        /*
-         * Set the TypeError like PyNumber_Index(o) would after trying
-         * the general case.
-         */
-        PyErr_Clear();
-    }
-
-    /*
-     * For backward compatibility check the number C-Api number protcol
-     * This should be removed up the finish label after deprecation.
-     */
-    if (Py_TYPE(o)->tp_as_number != NULL &&
-        Py_TYPE(o)->tp_as_number->nb_int != NULL) {
-        obj = Py_TYPE(o)->tp_as_number->nb_int(o);
-        if (obj == NULL) {
-            return -1;
-        }
- #if (NPY_SIZEOF_LONG < NPY_SIZEOF_INTP)
-        long_value = PyLong_AsLongLong(obj);
- #else
-        long_value = PyLong_AsLong(obj);
- #endif
-        Py_DECREF(obj);
-    }
-#if !defined(NPY_PY3K)
-    else if (Py_TYPE(o)->tp_as_number != NULL &&
-             Py_TYPE(o)->tp_as_number->nb_long != NULL) {
-        obj = Py_TYPE(o)->tp_as_number->nb_long(o);
-        if (obj == NULL) {
-            return -1;
-        }
-  #if (NPY_SIZEOF_LONG < NPY_SIZEOF_INTP)
-        long_value = PyLong_AsLongLong(obj);
-  #else
-        long_value = PyLong_AsLong(obj);
-  #endif
-        Py_DECREF(obj);
-    }
-#endif
-    else {
-        PyErr_SetString(PyExc_TypeError, msg);
+    if (obj == NULL) {
         return -1;
     }
-    /* Give a deprecation warning, unless there was already an error */
-    if (!error_converting(long_value)) {
-        if (DEPRECATE("using a non-integer number instead of an integer"
-                      " will result in an error in the future") < 0) {
-            return -1;
-        }
-    }
+#if (NPY_SIZEOF_LONG < NPY_SIZEOF_INTP)
+    long_value = PyLong_AsLongLong(obj);
+#else
+    long_value = PyLong_AsLong(obj);
+#endif
+    Py_DECREF(obj);
 
- finish:
     if (error_converting(long_value)) {
         err = PyErr_Occurred();
         /* Only replace TypeError's here, which are the normal errors. */
@@ -878,8 +962,9 @@ PyArray_PyIntAsIntp(PyObject *o)
         }
         return -1;
     }
+    goto overflow_check; /* silence unused warning */
 
- overflow_check:
+overflow_check:
 #if (NPY_SIZEOF_LONG < NPY_SIZEOF_INTP)
   #if (NPY_SIZEOF_LONGLONG > NPY_SIZEOF_INTP)
     if ((long_value < NPY_MIN_INTP) || (long_value > NPY_MAX_INTP)) {
@@ -900,6 +985,13 @@ PyArray_PyIntAsIntp(PyObject *o)
     return long_value;
 }
 
+/*NUMPY_API*/
+NPY_NO_EXPORT npy_intp
+PyArray_PyIntAsIntp(PyObject *o)
+{
+    return PyArray_PyIntAsIntp_ErrMsg(o, "an integer is required");
+}
+
 
 /*
  * PyArray_IntpFromIndexSequence
@@ -995,9 +1087,6 @@ NPY_NO_EXPORT int
 PyArray_TypestrConvert(int itemsize, int gentype)
 {
     int newtype = NPY_NOTYPE;
-    PyArray_Descr *temp;
-    const char *msg = "Specified size is invalid for this data type.\n"
-        "Size will be ignored in NumPy 1.7 but may throw an exception in future versions.";
 
     switch (gentype) {
         case NPY_GENBOOLLTR:
@@ -1114,6 +1203,7 @@ PyArray_TypestrConvert(int itemsize, int gentype)
             if (itemsize == 4 || itemsize == 8) {
                 int ret = 0;
                 if (evil_global_disable_warn_O4O8_flag) {
+                    /* 2012-02-04, 1.7, not sure when this can be removed */
                     ret = DEPRECATE("DType strings 'O4' and 'O8' are "
                             "deprecated because they are platform "
                             "specific. Use 'O' instead");
@@ -1151,31 +1241,6 @@ PyArray_TypestrConvert(int itemsize, int gentype)
             break;
     }
 
-    /*
-     * Raise deprecate warning if new type hasn't been
-     * set yet and size char is invalid.
-     * This should eventually be changed to an error in
-     * future NumPy versions.
-     */
-    if (newtype == NPY_NOTYPE) {
-        temp = PyArray_DescrFromType(gentype);
-        if (temp != NULL) {
-            if (temp->elsize != itemsize) {
-                if (DEPRECATE(msg) < 0) {
-                    Py_DECREF(temp);
-                    return -1;
-                }
-
-                newtype = gentype;
-            }
-            else {
-                newtype = gentype;
-            }
-
-            Py_DECREF(temp);
-        }
-    }
-
     return newtype;
 }
 
@@ -1185,7 +1250,7 @@ PyArray_TypestrConvert(int itemsize, int gentype)
   PyArray_IntTupleFromIntp
 */
 NPY_NO_EXPORT PyObject *
-PyArray_IntTupleFromIntp(int len, npy_intp *vals)
+PyArray_IntTupleFromIntp(int len, npy_intp const *vals)
 {
     int i;
     PyObject *intTuple = PyTuple_New(len);
@@ -1194,11 +1259,7 @@ PyArray_IntTupleFromIntp(int len, npy_intp *vals)
         goto fail;
     }
     for (i = 0; i < len; i++) {
-#if NPY_SIZEOF_INTP <= NPY_SIZEOF_LONG
-        PyObject *o = PyInt_FromLong((long) vals[i]);
-#else
-        PyObject *o = PyLong_FromLongLong((npy_longlong) vals[i]);
-#endif
+        PyObject *o = PyArray_PyIntFromIntp(vals[i]);
         if (!o) {
             Py_DECREF(intTuple);
             intTuple = NULL;
